#include "turboPmacController.h"
#include "asynMotorController.h"
#include "asynOctetSyncIO.h"
#include "turboPmacAxis.h"
#include <epicsExport.h>
#include <errlog.h>
#include <initHooks.h>
#include <iocsh.h>
#include <netinet/in.h>
#include <registryFunction.h>
#include <string.h>
#include <unistd.h>

/**
 * @brief Copy src into dst and replace all carriage returns with spaces. This
 * allows to print *dst with asynPrint.
 *
 *
 * @param dst                 Buffer for the modified string
 * @param src                 Original string
 */
void adjustResponseForPrint(char *dst, const char *src, size_t buf_length) {
    for (size_t i = 0; i < buf_length; i++) {
        if (src[i] == '\r') {
            dst[i] = ' ';
        } else {
            dst[i] = src[i];
        }
    }
}

turboPmacController::turboPmacController(const char *portName,
                                         const char *ipPortConfigName,
                                         int numAxes, double movingPollPeriod,
                                         double idlePollPeriod,
                                         double comTimeout, int numExtraParams)
    : sinqController(
          portName, ipPortConfigName, numAxes, movingPollPeriod, idlePollPeriod,
          /*
          The following parameter library entries are added in this driver:
          - REREAD_ENCODER_POSITION
          - READ_CONFIG
          */
          numExtraParams + NUM_turboPmac_DRIVER_PARAMS)

{

    // Initialization of local variables
    asynStatus status = asynSuccess;

    // Initialization of all member variables
    comTimeout_ = comTimeout;

    // Maximum allowed number of subsequent timeouts before the user is
    // informed.
    maxSubsequentTimeouts_ = 10;

    // =========================================================================
    // Create additional parameter library entries

    status = createParam("REREAD_ENCODER_POSITION", asynParamInt32,
                         &rereadEncoderPosition_);
    if (status != asynSuccess) {
        asynPrint(this->pasynUserSelf, ASYN_TRACE_ERROR,
                  "Controller \"%s\" => %s, line %d\nFATAL ERROR (creating a "
                  "parameter failed with %s).\nTerminating IOC",
                  portName, __PRETTY_FUNCTION__, __LINE__,
                  stringifyAsynStatus(status));
        exit(-1);
    }

    status = createParam("READ_CONFIG", asynParamInt32, &readConfig_);
    if (status != asynSuccess) {
        asynPrint(this->pasynUserSelf, ASYN_TRACE_ERROR,
                  "Controller \"%s\" => %s, line %d\nFATAL ERROR (creating a "
                  "parameter failed with %s).\nTerminating IOC",
                  portName, __PRETTY_FUNCTION__, __LINE__,
                  stringifyAsynStatus(status));
        exit(-1);
    }

    /*
    Define the end-of-string of a message coming from the device to EPICS.
    It is not necessary to append a terminator to outgoing messages, since
    the message length is encoded in the message header in the getSetResponse
    method.
    */
    const char *message_from_device =
        "\006"; // Hex-code for ACK (acknowledge) -> Each message from the MCU
                // is terminated by this value
    status = pasynOctetSyncIO->setInputEos(ipPortUser_, message_from_device,
                                           strlen(message_from_device));
    if (status != asynSuccess) {
        asynPrint(this->pasynUserSelf, ASYN_TRACE_ERROR,
                  "Controller \"%s\" => %s, line %d\nFATAL ERROR "
                  "(setting input EOS failed with %s).\nTerminating IOC",
                  portName, __PRETTY_FUNCTION__, __LINE__,
                  stringifyAsynStatus(status));
        pasynOctetSyncIO->disconnect(ipPortUser_);
        exit(-1);
    }

    status = callParamCallbacks();
    if (status != asynSuccess) {
        asynPrint(this->pasynUserSelf, ASYN_TRACE_ERROR,
                  "Controller \"%s\" => %s, line %d\nFATAL ERROR "
                  "(executing ParamLib callbacks failed "
                  "with %s).\nTerminating IOC",
                  portName, __PRETTY_FUNCTION__, __LINE__,
                  stringifyAsynStatus(status));
        pasynOctetSyncIO->disconnect(ipPortUser_);
        exit(-1);
    }
}

/*
Access one of the axes of the controller via the axis adress stored in asynUser.
If the axis does not exist or is not a Axis, a nullptr is returned and an
error is emitted.
*/
turboPmacAxis *turboPmacController::getTurboPmacAxis(asynUser *pasynUser) {
    asynMotorAxis *asynAxis = asynMotorController::getAxis(pasynUser);
    return dynamic_cast<turboPmacAxis *>(asynAxis);
}

/*
Access one of the axes of the controller via the axis index.
If the axis does not exist or is not a Axis, the function must return Null
*/
turboPmacAxis *turboPmacController::getTurboPmacAxis(int axisNo) {
    asynMotorAxis *asynAxis = asynMotorController::getAxis(axisNo);
    return dynamic_cast<turboPmacAxis *>(asynAxis);
}

asynStatus turboPmacController::writeRead(int axisNo, const char *command,
                                          char *response,
                                          int numExpectedResponses) {

    // Definition of local variables.
    asynStatus status = asynSuccess;
    asynStatus paramLibStatus = asynSuccess;
    asynStatus timeoutStatus = asynSuccess;
    // char fullCommand[MAXBUF_] = {0};
    char drvMessageText[MAXBUF_] = {0};
    char modResponse[MAXBUF_] = {0};
    int motorStatusProblem = 0;
    int numReceivedResponses = 0;

    /*
    asyn defines the following reasons for an end-of-message coming from the MCU
    (https://epics.anl.gov/modules/soft/asyn/R4-14/asynDriver.pdf, p. 28):
    0: Timeout
    1: Request count reached
    2: End of string detected -> In this driver, this is the "normal" case
    4: End indicator detected
    Combinations of reasons are also possible, e.g. eomReason = 5 would mean
    that both the request count was reached and an end indicator was detected.
     */
    int eomReason = 0;

    // Number of bytes of the outgoing message (which is command + the
    // end-of-string terminator defined in the constructor)
    size_t nbytesOut = 0;

    // Number of bytes of the incoming message (which is response + the
    // end-of-string terminator defined in the constructor)
    size_t nbytesIn = 0;

    // =========================================================================

    turboPmacAxis *axis = getTurboPmacAxis(axisNo);
    if (axis == nullptr) {
        // We already did the error logging directly in getAxis
        return asynError;
    }
    const size_t commandLength = strlen(command);

    // NOT NEEDED ANYMORE DUE TO THE LOW LEVEL DRIVER
    // /*
    // The message protocol of the turboPmac used at PSI looks as follows (all
    // characters immediately following each other without a newline):
    // 0x40 (ASCII value of @) -> Request for download
    // 0xBF (ASCII value of ¿) -> Select mode "get_response"
    // 0x00 (ASCII value of 0)
    // 0x00 (ASCII value of 0)
    // 0x00 (ASCII value of 0)
    // 0x00 (ASCII value of 0)
    // 0x00 (ASCII value of 0)
    // [message length in network byte order] -> Use the htons function for this
    // value [Actual message] It is not necessary to append a terminator, since
    // this protocol encodes the message length at the beginning. See Turbo PMAC
    // User Manual, page 418 in VR_PMAC_GETRESPONSE
    // x0D (ASCII value of carriage return) -> The controller needs a carriage
    // return at the end of a "send" command (a command were we transmit data
    // via
    // =). For "request" commands (e.g. read status or position), this is not
    // necessary, but it doesn't hurt either, therefore we always add a carriage
    // return.

    // The message has to be build manually into the buffer fullCommand, since
    // it contains NULL terminators in its middle, therefore the string
    // manipulation methods of C don't work.
    // */

    //
    // const int offset = 9;

    // // Positions 2 to 6 must have the value 0. Since fullCommand is
    // initialized
    // // as an array of zeros, we don't need to set these bits manually.
    // fullCommand[0] = '\x40';
    // fullCommand[1] = '\xBF';

    // // The size of size_t is platform dependant (pointers-sized), while htons
    // // needs an unsigned int. The byte order is then converted from host to
    // // network order. The offset "+1" is for the carriage return.
    // u_int16_t len = htons(static_cast<u_int16_t>(commandLength + 1));

    // // Split up into the upper and the lower byte
    // fullCommand[7] = (char)(len >> 8);   // Shift the 8 higher bits to the
    // right fullCommand[8] = (char)(len & 0xFF); // Mask the higher bits

    // // Write the actual command behind the protocol
    // for (size_t i = 0; i < commandLength; i++) {
    //     fullCommand[i + offset] = command[i];
    // }
    // fullCommand[offset + commandLength] = '\x0D';

    // // +1 for the carriage return.
    // const size_t fullComandLength = offset + commandLength + 1;

    /*
    We use separated write and read commands here, not the combined writeRead
    method, because the latter is actually a flushWriteRead (see
    https://epics.anl.gov/modules/soft/asyn/R4-14/asynDriver.pdf, p. 31) ->
    Calls the flush command, then the write command, then the read command.
    The flush itself reads repeatedly from the MCU until no messages are there
    anymore. (The Diamond Light Source driver first send a PMAC flush command
    and then does the same as the asyn flush). We don't want this behaviour.
    (https://www.slac.stanford.edu/grp/lcls/controls/global/doc/epics-modules/R3-14-12/asyn/asyn-R4-18-lcls2/asyn/interfaces/asynOctetBase.c)

    If a timeout occurs during writing or reading, inform the user that we're
    trying to reconnect. If the problem persists, ask them to call the support
    */

    status = pasynOctetSyncIO->write(ipPortUser_, command, commandLength,
                                     comTimeout_, &nbytesOut);

    msgPrintControlKey writeKey =
        msgPrintControlKey(portName, axisNo, __PRETTY_FUNCTION__, __LINE__);

    if (status == asynTimeout) {

        if (msgPrintControl_.shouldBePrinted(writeKey, true, pasynUserSelf)) {
            asynPrint(
                this->pasynUserSelf, ASYN_TRACE_ERROR,
                "Controller \"%s\", axis %d => %s, line %d\nTimeout while "
                "writing to the MCU.%s\n",
                portName, axisNo, __PRETTY_FUNCTION__, __LINE__,
                msgPrintControl_.getSuffix());
        }

        timeoutStatus = checkComTimeoutWatchdog(axisNo, drvMessageText,
                                                sizeof(drvMessageText));

        int timeoutCounter = 0;
        while (1) {
            checkMaxSubsequentTimeouts(timeoutCounter, axis);
            timeoutCounter += 1;

            status = pasynOctetSyncIO->write(
                ipPortUser_, command, commandLength, comTimeout_, &nbytesOut);
            if (status != asynTimeout) {
                asynPrint(this->pasynUserSelf, ASYN_TRACE_ERROR,
                          "Controller \"%s\", axis %d  => %s, line "
                          "%d\nReconnected after write timeout\n",
                          portName, axisNo, __PRETTY_FUNCTION__, __LINE__);
                break;
            }
        }
    } else if (status != asynSuccess) {
        if (msgPrintControl_.shouldBePrinted(writeKey, true, pasynUserSelf)) {
            asynPrint(
                this->pasynUserSelf, ASYN_TRACE_ERROR,
                "Controller \"%s\", axis %d => %s, line %d\nError %s while "
                "writing to the controller.%s\n",
                portName, axisNo, __PRETTY_FUNCTION__, __LINE__,
                stringifyAsynStatus(status), msgPrintControl_.getSuffix());
        }
    } else {
        msgPrintControl_.resetCount(writeKey, pasynUserSelf);
    }

    // Read the response from the MCU buffer
    status = pasynOctetSyncIO->read(ipPortUser_, response, MAXBUF_, comTimeout_,
                                    &nbytesIn, &eomReason);

    msgPrintControlKey readKey =
        msgPrintControlKey(portName, axisNo, __PRETTY_FUNCTION__, __LINE__);

    if (status == asynTimeout) {

        if (msgPrintControl_.shouldBePrinted(readKey, true, pasynUserSelf)) {
            asynPrint(
                this->pasynUserSelf, ASYN_TRACE_ERROR,
                "Controller \"%s\", axis %d => %s, line %d\nTimeout while "
                "reading from the MCU.%s\n",
                portName, axisNo, __PRETTY_FUNCTION__, __LINE__,
                msgPrintControl_.getSuffix());
        }

        // Add this event to the back of the timeout event counter
        timeoutStatus = checkComTimeoutWatchdog(axisNo, drvMessageText,
                                                sizeof(drvMessageText));

        int timeoutCounter = 0;
        while (1) {
            checkMaxSubsequentTimeouts(timeoutCounter, axis);
            timeoutCounter += 1;

            status = pasynOctetSyncIO->read(ipPortUser_, response, MAXBUF_,
                                            comTimeout_, &nbytesIn, &eomReason);
            if (status != asynTimeout) {
                asynPrint(this->pasynUserSelf, ASYN_TRACE_ERROR,
                          "Controller \"%s\", axis %d  => %s, line "
                          "%d\nReconnected after read timeout\n",
                          portName, axisNo, __PRETTY_FUNCTION__, __LINE__);
                break;
            }
        }
    } else if (status != asynSuccess) {
        if (msgPrintControl_.shouldBePrinted(readKey, true, pasynUserSelf)) {
            asynPrint(
                this->pasynUserSelf, ASYN_TRACE_ERROR,
                "Controller \"%s\", axis %d => %s, line %d\nError %s while "
                "reading from the controller.%s\n",
                portName, axisNo, __PRETTY_FUNCTION__, __LINE__,
                stringifyAsynStatus(status), msgPrintControl_.getSuffix());
        }
    } else {
        msgPrintControl_.resetCount(readKey, pasynUserSelf);
    }

    if (timeoutStatus == asynError) {
        status = asynError;
    }

    // The message should only ever terminate due to reason 2
    msgPrintControlKey terminateKey =
        msgPrintControlKey(portName, axisNo, __PRETTY_FUNCTION__, __LINE__);
    if (eomReason != 2) {
        status = asynError;

        snprintf(drvMessageText, sizeof(drvMessageText),
                 "Terminated message due to reason %d (should be 2).",
                 eomReason);

        if (msgPrintControl_.shouldBePrinted(terminateKey, true,
                                             pasynUserSelf)) {
            asynPrint(this->pasynUserSelf, ASYN_TRACE_ERROR,
                      "Controller \"%s\", axis %d => %s, line %d\nMessage "
                      "terminated due to reason %i.%s\n",
                      portName, axisNo, __PRETTY_FUNCTION__, __LINE__,
                      eomReason, msgPrintControl_.getSuffix());
        }
    } else {
        msgPrintControl_.resetCount(terminateKey, pasynUserSelf);
    }

    /*
    Calculate the number of received responses by counting the number of
    carriage returns "\r" in the response.
    */
    for (size_t i = 0; i < strlen(response); i++) {
        if (response[i] == '\r') {
            numReceivedResponses++;
        }
    }
    msgPrintControlKey numResponsesKey =
        msgPrintControlKey(portName, axisNo, __PRETTY_FUNCTION__, __LINE__);
    if (numExpectedResponses != numReceivedResponses) {
        adjustResponseForPrint(modResponse, response, MAXBUF_);

        if (msgPrintControl_.shouldBePrinted(numResponsesKey, true,
                                             pasynUserSelf)) {
            asynPrint(
                this->pasynUserSelf, ASYN_TRACE_ERROR,
                "Controller \"%s\", axis %d => %s, line %d\nUnexpected "
                "response '%s' (carriage returns are replaced with spaces) "
                "for command %s.%s\n",
                portName, axisNo, __PRETTY_FUNCTION__, __LINE__, modResponse,
                command, msgPrintControl_.getSuffix());
        }

        snprintf(drvMessageText, sizeof(drvMessageText),
                 "Received unexpected response '%s' (carriage returns "
                 "are replaced with spaces) for command %s. "
                 "Please call the support",
                 modResponse, command);
        status = asynError;
    } else {
        msgPrintControl_.resetCount(numResponsesKey, pasynUserSelf);
    }

    // Create custom error messages for different failure modes, if no error
    // message has been set yet
    if (strlen(drvMessageText) == 0) {
        switch (status) {
        case asynSuccess:
            break; // Communicate nothing
        case asynTimeout:
            snprintf(drvMessageText, sizeof(drvMessageText),
                     "connection timeout for axis %d", axisNo);
            break;
        case asynDisconnected:
            snprintf(drvMessageText, sizeof(drvMessageText),
                     "axis is not connected");
            break;
        case asynDisabled:
            snprintf(drvMessageText, sizeof(drvMessageText),
                     "axis is disabled");
            break;
        default:
            snprintf(drvMessageText, sizeof(drvMessageText),
                     "Communication failed (%s)", stringifyAsynStatus(status));
            break;
        }
    }

    // Log the overall status (communication successfull or not)
    if (status == asynSuccess) {
        paramLibStatus = axis->setIntegerParam(this->motorStatusCommsError_, 0);
    } else {
        // Check if the axis already is in an error communication mode. If
        // it is not, upstream the error. This is done to avoid "flooding"
        // the user with different error messages if more than one error
        // ocurred before an error-free communication
        paramLibStatus =
            getIntegerParam(axisNo, motorStatusProblem_, &motorStatusProblem);
        if (paramLibStatus != asynSuccess) {
            return paramLibAccessFailed(paramLibStatus, "motorStatusProblem_",
                                        axisNo, __PRETTY_FUNCTION__, __LINE__);
        }

        if (motorStatusProblem == 0) {
            paramLibStatus =
                axis->setStringParam(motorMessageText_, drvMessageText);
            if (paramLibStatus != asynSuccess) {
                return paramLibAccessFailed(paramLibStatus, "motorMessageText_",
                                            axisNo, __PRETTY_FUNCTION__,
                                            __LINE__);
            }

            paramLibStatus = axis->setIntegerParam(motorStatusProblem_, 1);
            if (paramLibStatus != asynSuccess) {
                return paramLibAccessFailed(paramLibStatus,
                                            "motorStatusProblem", axisNo,
                                            __PRETTY_FUNCTION__, __LINE__);
            }

            paramLibStatus = axis->setIntegerParam(motorStatusProblem_, 1);
            if (paramLibStatus != asynSuccess) {
                return paramLibAccessFailed(paramLibStatus,
                                            "motorStatusCommsError_", axisNo,
                                            __PRETTY_FUNCTION__, __LINE__);
            }
        }
    }
    return status;
}

asynStatus turboPmacController::writeInt32(asynUser *pasynUser,
                                           epicsInt32 value) {
    int function = pasynUser->reason;

    // =====================================================================

    turboPmacAxis *axis = getTurboPmacAxis(pasynUser);

    // Handle custom PVs
    if (function == rereadEncoderPosition_) {
        return axis->rereadEncoder();
    } else if (function == readConfig_) {
        return axis->init();
    } else {
        return sinqController::writeInt32(pasynUser, value);
    }
}

asynStatus turboPmacController::couldNotParseResponse(const char *command,
                                                      const char *response,
                                                      int axisNo,
                                                      const char *functionName,
                                                      int lineNumber) {
    char modifiedResponse[MAXBUF_] = {0};
    adjustResponseForPrint(modifiedResponse, response, MAXBUF_);
    return sinqController::couldNotParseResponse(
        command, modifiedResponse, axisNo, functionName, lineNumber);
}

/*************************************************************************************/
/** The following functions are C-wrappers, and can be called directly from
 * iocsh */

extern "C" {

/*
C wrapper for the controller constructor. Please refer to the
turboPmacController constructor documentation.
*/
asynStatus turboPmacCreateController(const char *portName,
                                     const char *ipPortConfigName, int numAxes,
                                     double movingPollPeriod,
                                     double idlePollPeriod, double comTimeout) {
    /*
    We create a new instance of the controller, using the "new" keyword to
    allocate it on the heap while avoiding RAII.
    https://github.com/epics-modules/motor/blob/master/motorApp/MotorSrc/asynMotorController.cpp
    https://github.com/epics-modules/asyn/blob/master/asyn/asynPortDriver/asynPortDriver.cpp

    The created object is registered in EPICS in its constructor and can
    safely be "leaked" here.
    */
#pragma GCC diagnostic ignored "-Wunused-but-set-variable"
#pragma GCC diagnostic ignored "-Wunused-variable"
    turboPmacController *pController =
        new turboPmacController(portName, ipPortConfigName, numAxes,
                                movingPollPeriod, idlePollPeriod, comTimeout);

    return asynSuccess;
}

/*
Define name and type of the arguments for the CreateController function
in the iocsh. This is done by creating structs with the argument names and
types and then providing "factory" functions
(configCreateControllerCallFunc). These factory functions are used to
register the constructors during compilation.
*/
static const iocshArg CreateControllerArg0 = {"Controller name (e.g. mcu1)",
                                              iocshArgString};
static const iocshArg CreateControllerArg1 = {"Asyn IP port name (e.g. pmcu1)",
                                              iocshArgString};
static const iocshArg CreateControllerArg2 = {"Number of axes", iocshArgInt};
static const iocshArg CreateControllerArg3 = {"Moving poll rate (s)",
                                              iocshArgDouble};
static const iocshArg CreateControllerArg4 = {"Idle poll rate (s)",
                                              iocshArgDouble};
static const iocshArg CreateControllerArg5 = {"Communication timeout (s)",
                                              iocshArgDouble};
static const iocshArg *const CreateControllerArgs[] = {
    &CreateControllerArg0, &CreateControllerArg1, &CreateControllerArg2,
    &CreateControllerArg3, &CreateControllerArg4, &CreateControllerArg5};
static const iocshFuncDef configTurboPmacCreateController = {
    "turboPmacController", 6, CreateControllerArgs};
static void configTurboPmacCreateControllerCallFunc(const iocshArgBuf *args) {
    turboPmacCreateController(args[0].sval, args[1].sval, args[2].ival,
                              args[3].dval, args[4].dval, args[5].dval);
}

// This function is made known to EPICS in turboPmac.dbd and is called by
// EPICS in order to register both functions in the IOC shell
static void turboPmacControllerRegister(void) {
    iocshRegister(&configTurboPmacCreateController,
                  configTurboPmacCreateControllerCallFunc);
}
epicsExportRegistrar(turboPmacControllerRegister);

} // extern "C"
